@@ -33,16 +33,18 @@ class RuleParseError(Exception):
 #
 
 class RuleSet(object):
-    def __init__(self, rules, cache_generation=True, cache_matching=True):
+    def __init__(self, rules=None, cache_generation=True, cache_matching=True):
+        if rules is not None and not rules:
+            raise Exception('No rules specified')
         self.to_vars_converter = convertAndCache(
             [r.to_vars for r in rules],
             cache=cache_matching,
-            type='match', 
+            convert_type='match', 
         )
         self.to_url_converter = convertAndCache(
             [r.to_url for r in rules],
             cache=cache_generation,
-            type='generate', 
+            convert_type='generate', 
         )
 
     def match_url(self, url, script=None):
@@ -55,7 +57,8 @@ class RuleSet(object):
                 )
             url_parts['script'] = script
             url_parts['path'] = url_parts['path'][len(script)+1:]
-        return self.match(url_parts)
+        result = self.match(url_parts)
+        return result
     
     def match(self, url_parts):
         log.debug('Converting URL parts %r to routing variables', url_parts)
@@ -66,8 +69,11 @@ class RuleSet(object):
         rv = dict([(unicode(k),unicode(v)) for k,v in routing_vars.items()])
         conversion = self.generate(rv, default_url_parts)
         if conversion.successful:
-            d.update(conversion.result)
-            return build_url(**d)
+            res = conversion.result
+            d.update(res.url_parts)
+            # @@@ Should really use a post converter
+            result = Conversion(AttributeDict(url=build_url(**d), extra=res.extra)).perform(noConversion())
+            return result
         raise Exception('No matching generation rule could be found')
 
     def generate(self, routing_vars, default_url_parts=None):
@@ -82,35 +88,35 @@ class RuleSet(object):
 # Cache
 #
 
-def convertAndCache(converters, cache=True, type=None):
+def convertAndCache(converters, cache=True, convert_type=None):
     if cache:
         cached = {}
     def convertAndCache_converter(conversion, service=None):
         parts = conversion.value
         if cache:
-            if type == 'generate':
+            if convert_type == 'generate':
                 a = parts['vars'].items()
                 b = parts['current'].items()
                 a.sort()
                 b.sort()
                 p = (tuple(a), tuple(b),)
-            elif type == 'match':
+            elif convert_type == 'match':
                 p = parts.items()
                 p.sort()
                 p = (tuple(p), None)
             else:
-                raise URLConvertError('Unknown type for caching %r'%type)
+                raise URLConvertError('Unknown convert_type for caching %r'%convert_type)
             if p in cached:
-                log.info(
-                    'Converting from cache: %r, type: %r, result: %r',
+                log.debug(
+                    'Converting from cache: %r, convert_type: %r, result: %r',
                     parts, 
-                    type,
+                    convert_type,
                     cached[p],
                 )
                 conversion.result = cached[p]
                 return
 
-        log.debug('Not cached: %r, type: %r', parts, type)
+        log.debug('Not cached: %r, convert_type: %r', parts, convert_type)
         log.debug('%s converters available', len(converters))
         log.debug('Trying each rule in turn...')
         conversion.perform(
@@ -121,19 +127,20 @@ def convertAndCache(converters, cache=True, type=None):
             )
         )
         if conversion.successful:
+            set_result(conversion, AttributeDict(conversion.result))
             if cache:
                 cached[p] = conversion.result
                 log.info(
-                    'Converting and adding to cache: %r, type: %r, result: %r',
+                    'Converting and adding to cache: %r, convert_type: %r, result: %r',
                     parts, 
-                    type,
+                    convert_type,
                     cached[p],
                 )
             else:
-                log.info(
-                    'Converting, not caching: %r, type: %r, result: %r',
+                log.debug(
+                    'Converting, not caching: %r, convert_type: %r, result: %r',
                     parts, 
-                    type,
+                    convert_type,
                     conversion.result,
                 )
         else:
@@ -183,7 +190,7 @@ def urlToParts():
             host = rest[0]
             port = None
             # No port was explicitly specified so make a best guess:
-            environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+            environ = service and hasattr(service, 'environ') and service.environ or None
             if environ:
                 port = unicode(
                     environ.get(
@@ -622,7 +629,7 @@ def matchDynamicPath(part, path_part):
 # Creating converters from a rule
 #
 
-def rule(rule, add=None):
+def rule(rule, add=None, extra=None):
     if not isinstance(rule, unicode):
         raise URLConvertError(
             'Expected the rule to be specified as a Unicode string'
@@ -731,31 +738,40 @@ def rule(rule, add=None):
     # Finally, create a converter capable of parsing all this data
     post_converters = [
         toDictionary(match), 
-        mergeParts()
+        mergeParts(),
     ]
     if add is not None:
         post_converters.append(addExtras(add))
+    post_converters.append(extraVariables(extra))
     to_vars = chainConverters(
         missingKey(match.keys(), 'match'),
         chainPostConverters(*post_converters)
     )
     to_url = chainPostConverters(
-           chainConverters(
-                removeExtras(add or {}),
-                # This is expecting a dictionary, we have a tuple, and want each 
-                # of the generate converters called for each tuple.
-                # ie: We need a for_each with a key for each.
-                tryEach(
-                    generate.values(),
-                    stop_on_first_result=False, 
-                    stop_on_first_error=True, 
-                    children_keys=generate.keys()
-                ),
+       chainConverters(
+            removeExtras(add or {}),
+            # This is expecting a dictionary, we have a tuple, and want each 
+            # of the generate converters called for each tuple.
+            # ie: We need a for_each with a key for each.
+            tryEach(
+                generate.values(),
+                stop_on_first_result=False, 
+                stop_on_first_error=True, 
+                children_keys=generate.keys()
             ),
-            joinVars(add or {}),
-        )
+        ),
+        joinVars(add or {}),
+        extraVariables(extra, 'url_parts'),
+    )
     return AttributeDict(dict(to_vars=to_vars, to_url=to_url))
 
+def extraVariables(extra, key='vars'):
+    def extraVariables_post_converter(conversion, state):
+        if conversion.successful:
+            vars = conversion.result
+            set_result(conversion, AttributeDict(extra=extra, **{key:vars}))
+    return extraVariables_post_converter
+
 #
 # Rule handling helpers
 #
@@ -898,7 +914,7 @@ def matchScheme(make_lowercase=True, allowed_values=None):
             conversion.result = scheme
     return matchScheme_converter
 
-domain_label_regex = re.compile('[a-zA-Z0-9]([a-zA-Z0-9\-]*)[a-zA-Z0-9]')
+domain_label_regex = re.compile('([0-9]+) || [a-zA-Z0-9]([a-zA-Z0-9\-]*)[a-zA-Z0-9]')
 
 def matchHost():
     """\
@@ -1150,7 +1166,7 @@ def extract_scheme(service, converter=None):
        hasattr(service.config.server, 'display_scheme'):
         scheme = service.config.app.display_scheme
     else:
-        environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+        environ = service and hasattr(service, 'environ') and service.environ or None
         if environ:
             scheme = environ.get('wsgi.url_scheme')
             if not scheme:
@@ -1188,7 +1204,7 @@ def extract_host(service, converter=None):
        hasattr(service.config.server, 'display_host'):
         host = service.config.server.display_host
     else:
-        environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+        environ = service and hasattr(service, 'environ') and service.environ or None
         if environ:
             if environ.has_key('X_FORWARDED_FOR'):
                 host = environ['X_FORWARDED_FOR'].split(',')[0].strip()
@@ -1219,7 +1235,7 @@ def extract_port(service, converter=None):
     """
     if converter is None:
         converter = decode_port
-    environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+    environ = service and hasattr(service, 'environ') and service.environ or None
     if environ:
         port = str(environ.get('X_PORT_FORWARDED_FOR', environ.get('SERVER_PORT')))
     if not port:
@@ -1257,13 +1273,13 @@ def extract_path(service, converter=None):
     """
     if converter is None:
         converter = decode_path
-    environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+    environ = service and hasattr(service, 'environ') and service.environ or None
     if environ:
         path = environ.get('PATH_INFO')
         if path is None:
             return None
         if not path.startswith('/'):
-            raise URLConvertError("Expected the PATH_INFO to start with '/'")
+            raise URLConvertError("Expected the PATH_INFO %r to start with '/'"%(path))
         path = path[1:]
     return Conversion(path).perform(converter).result
 
@@ -1281,15 +1297,21 @@ def extract_script(service, converter=None):
     """
     if converter is None:
         converter = decode_script
-    environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+    environ = service and hasattr(service, 'environ') and \
+       service.environ or None
     if environ:
         script = environ.get('SCRIPT_NAME')
         if script is None:
             return None
-        # It seems that the SCRIPT_NAME can be '' which is irritating.
         if script.startswith('/'):
-            #raise URLConvertError("Expected the SCRIPT_NAME to start with '/' or be empty")
             script = script[1:]
+        if script.endswith('/'):
+            raise URLConvertError(
+                "The SCRIPT_NAME %r cannot end with a '/' character" %(
+                    script 
+                )
+            )
+            #script = script[:-1]
     return Conversion(script).perform(converter).result
 
 def extract_query(service, converter=None):
@@ -1306,7 +1328,8 @@ def extract_query(service, converter=None):
     """
     if converter is None:
         converter = decode_query
-    environ = service and hasattr(service, 'wsgi') and service.wsgi.environ or None
+    environ = service and hasattr(service, 'environ') and \
+       service.environ or None
     if environ:
         query = environ.get('QUERY_STRING')
     if query is None:
